#!/usr/bin/perl

################################################################################################################################
#  Authors: Svetlana Kiritchenko, Peter Turney
#  Information and Communications Technologies / Technologies de l'information et des communications
#  National Research Council Canada /Conseil national de recherches Canada
#  
#  Description: calculates real-valued scores for items annotated with Best-Worst Scaling method;
#               uses the simple counting procedure: (#Best - #Worst)/n. 
#
#  Usage: get-scores-from-BWS-annotations-counting.pl <file-annotations>
#    <file-annotations> is a CSV (comma delimitted) file with item tuples and Best-Worst annotations.
#
#  Output: items (and if present, their IDs) with scores (one item per line) 
#    The output is written into STDOUT. 
#  
#  To run the script, you need the package Text::CSV to be installed.
#
#  Version: 1.1
#  Last modified: August 24, 2017
#
#################################################################################################################################

use warnings;
use strict;
use utf8;
binmode (STDIN,  ":utf8");
binmode (STDOUT, ":utf8");

use Text::CSV;


#################################################################################################################################
# Names of the columns in the annotation data.
# PLEASE MODIFY AS NEEDED!
#
# @item_column_names lists the names of the columns that contain items
# @itemID_column_names lists the names of the columns that contain item IDs 
# @ann_column_names lists the names of the columns that contain Best and Worst annotations (e.i., items from a tuple chosen as Best and Worst)
#
# There can be more or less than 4 item columns. 
# The number of columns for items and the number of columns for item IDs have to be the same. (The ID in the first column of @itemID_column_names corresponds to the item in the first column of @item_column_names.)
# The array @itemID_column_names may be empty (if no item IDs are stored).
#################################################################################################################################

my @item_column_names = ("Item1", "Item2", "Item3", "Item4");
my @itemID_column_names = ("ItemID1", "ItemID2", "ItemID3", "ItemID4");
my @ann_column_names = ("BestItem", "WorstItem");

# check if the number of columns for items and the number of columns for item IDs are the same
if((scalar(@itemID_column_names) > 0) && (scalar(@item_column_names) != scalar(@itemID_column_names))) {
	print STDERR "ERROR: The number of columns for items (\@item_column_names) and for item IDs (\@itemID_column_names) have to be the same.\n";
	exit();
}

#################################################################################################################################


die "Usage: get-scores-from-BWS-annotations-counting.pl <file-annotations>\n" if(@ARGV < 1);

# file with the BWS annotations
my $file_annotations = $ARGV[0];


# read the Best-Worst annotation file
print STDERR "Reading the annotation file $file_annotations ...\n";

my %count_item = (); my %count_best = (); my %count_worst = ();

my $csv = Text::CSV->new ( { binary => 1 } )  # should set binary attribute.
                 or die "Cannot use CSV: ".Text::CSV->error_diag ();
 
open my $fh, "<", $file_annotations or die "Cannot open the annotation file $file_annotations: $!";

# read the the first line with the column names
# check if all @column_names can be found
my $column_names_row = $csv->getline($fh);
my %data_column_names = ();
foreach my $col (@{$column_names_row}) {
	$data_column_names{$col} = 1;
}

foreach my $col (@item_column_names, @itemID_column_names, @ann_column_names) {
	if(!defined $data_column_names{$col}) {
		print STDERR "ERROR: Cannot find a column named $col.\n";
		exit();
	}
}

$csv->column_names ($column_names_row);

# read the annotations line by line
my %item_IDs = ();
my $line_num = 1;
while (my $data = $csv->getline_hr($fh)) {
	
	# read the items (and their IDs)
	my %items_in_tuple = ();
	for(my $i = 0; $i < @item_column_names; $i++) {
		my $item = $data->{$item_column_names[$i]};
		
		# if item IDs are present, store item-itemID pairs
		if(@itemID_column_names) {
			my $itemID = $data->{$itemID_column_names[$i]};
			if((defined $item_IDs{$item}) && ($item_IDs{$item} ne $itemID)) {
				print STDERR "ERROR: Different IDs for item ($item): $item_IDs{$item} and $itemID\n";
				exit();
			}
			$item_IDs{$item} = $itemID;
		}
		$count_item{$item}++;
		$items_in_tuple{$item} = 1;
	}

	# read the Best and Worst items
	my $best_item = $data->{$ann_column_names[0]};
	my $worst_item = $data->{$ann_column_names[1]};
	
	# check if the Best and Worst items are among the tuple items
	if(!defined $items_in_tuple{$best_item}) {
		print STDERR "ERROR: Illegible annotation for the Best item in line $line_num.\n";
		exit();
	}
	if(!defined $items_in_tuple{$worst_item}) {
		print STDERR "ERROR: Illegible annotation for the Worst item in line $line_num.\n";
		exit();
	}

	# check if the Best and Worst items are the same
	if($best_item eq $worst_item) {
		print STDERR "WARNING: Annotations for the Best and Worst items are identical in line $line_num.\n";
	}
	
	$count_best{$best_item}++;
	$count_worst{$worst_item}++;

	$line_num++;  
}	
	
close($fh);

print STDERR "Read ".($line_num - 1)." annotations.\n";
print STDERR "Found ".scalar(keys %count_item)." unique items.\n";


# calculating the scores for the items
my %scores = ();
foreach my $item (keys %count_item) {
	if(!defined $count_best{$item}) { $count_best{$item} = 0; }
	if(!defined $count_worst{$item}) { $count_worst{$item} = 0; }
	
	$scores{$item} = ($count_best{$item} - $count_worst{$item}) / $count_item{$item};
}

# writing the scores to STDOUT
print STDERR "\nWriting the scores to STDOUT ...\n";

foreach my $item (sort {$scores{$b} <=> $scores{$a}} keys %scores) {
	if(@itemID_column_names) {
		print $item_IDs{$item}."\t";
	}
	printf "%s\t%.3f\n", $item, $scores{$item};
}

print STDERR "Finished.\n";



